# -*- coding: utf-8 -*-
"""
/***************************************************************************
 Plugin
                                 A QGIS plugin
 Generete isochrones
 Generated by Plugin Builder: http://g-sherman.github.io/Qgis-Plugin-Builder/
                              -------------------
        begin                : 2022-03-01
        git sha              : $Format:%H$
        copyright            : (C) 2022 by StraTopo
        email                : c.vaessens@stratopo.nl
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
"""
from distutils.log import error
from importlib import resources

from qgis.PyQt.QtCore import QSettings, QTranslator, QCoreApplication, QVariant
from qgis.PyQt.QtGui import QIcon
from qgis.PyQt.QtWidgets import QAction

from qgis.core import QgsVectorLayer, QgsFeature, QgsPointXY, QgsGeometry, QgsProject, QgsField, QgsWkbTypes

# Initialize Qt resources from file resources.py
from .resources.resources import *
# Import the code for the dialog
from .ui.isochrones_dialog import IsochronesDialog
from .ui.configure_dialog import ConfigureDialog
from .ui.bod_dialog import BodDialog
import os.path
import re
import requests

from .tools.resources import plugin_name
from .tools.settings import get_setting, set_setting

class Plugin:

    apiUrl = "https://api.stratopo.nl"

    """QGIS Plugin Implementation."""

    def __init__(self, iface):
        """Constructor.

        :param iface: An interface instance that will be passed to this class
            which provides the hook by which you can manipulate the QGIS
            application at run time.
        :type iface: QgsInterface
        """
        # Save reference to the QGIS interface
        self.iface = iface
        # initialize plugin directory
        self.plugin_dir = os.path.dirname(__file__)
        # initialize locale
        locale = QSettings().value('locale/userLocale')[0:2]
        locale_path = os.path.join(
            self.plugin_dir,
            'i18n',
            'Plugin_{}.qm'.format(locale))

        if os.path.exists(locale_path):
            self.translator = QTranslator()
            self.translator.load(locale_path)
            QCoreApplication.installTranslator(self.translator)

        # Declare instance attributes
        self.actions = []
        self.menu = plugin_name()


    # noinspection PyMethodMayBeStatic
    def tr(self, message):
        """Get the translation for a string using Qt translation API.

        We implement this ourselves since we do not inherit QObject.

        :param message: String for translation.
        :type message: str, QString

        :returns: Translated version of message.
        :rtype: QString
        """
        # noinspection PyTypeChecker,PyArgumentList,PyCallByClass
        return QCoreApplication.translate('Plugin', message)


    def add_action(
        self,
        icon_path,
        text,
        callback,
        enabled_flag=True,
        add_to_menu=True,
        add_to_toolbar=True,
        status_tip=None,
        whats_this=None,
        parent=None):
        """Add a toolbar icon to the toolbar.

        :param icon_path: Path to the icon for this action. Can be a resource
            path (e.g. ':/plugins/foo/bar.png') or a normal file system path.
        :type icon_path: str

        :param text: Text that should be shown in menu items for this action.
        :type text: str

        :param callback: Function to be called when the action is triggered.
        :type callback: function

        :param enabled_flag: A flag indicating if the action should be enabled
            by default. Defaults to True.
        :type enabled_flag: bool

        :param add_to_menu: Flag indicating whether the action should also
            be added to the menu. Defaults to True.
        :type add_to_menu: bool

        :param add_to_toolbar: Flag indicating whether the action should also
            be added to the toolbar. Defaults to True.
        :type add_to_toolbar: bool

        :param status_tip: Optional text to show in a popup when mouse pointer
            hovers over the action.
        :type status_tip: str

        :param parent: Parent widget for the new action. Defaults None.
        :type parent: QWidget

        :param whats_this: Optional text to show in the status bar when the
            mouse pointer hovers over the action.

        :returns: The action that was created. Note that the action is also
            added to self.actions list.
        :rtype: QAction
        """

        icon = QIcon(icon_path)
        action = QAction(icon, text, parent)
        action.triggered.connect(callback)
        action.setEnabled(enabled_flag)

        if status_tip is not None:
            action.setStatusTip(status_tip)

        if whats_this is not None:
            action.setWhatsThis(whats_this)

        if add_to_toolbar:
            # Adds plugin icon to Plugins toolbar
            self.iface.addToolBarIcon(action)

        if add_to_menu:
            self.iface.addPluginToMenu(
                self.menu,
                action)

        self.actions.append(action)

        return action

    def initGui(self):
        """Create the menu entries and toolbar icons inside the QGIS GUI."""

        self.add_action(
            ':/resources/images/IsochronesIcon.png',
            text='SGS Isochrones',
            callback=self.showIsochonesWindow,
            parent=self.iface.mainWindow())

        self.add_action(
            ':/resources/images/ConfigureIcon.png',
            text='SGS Configure',
            callback=self.showConfigurationWindow,
            parent=self.iface.mainWindow())

        self.add_action(
            ':/resources/images/BodIcon.png',
            text='SGS BOD',
            callback=self.showBodWindow,
            parent=self.iface.mainWindow())

        # icon_path = ':/plugins/stratopo_isochrones/icon.png'
        # self.action1 = QAction(QIcon(icon_path), u"StraTopo Isochrones", self.iface.mainWindow())
        # self.action1.triggered.connect(self.run)
        # self.action2 = QAction(QIcon(icon_path), u"Configure", self.iface.mainWindow())

        # self.popupMenu = QMenu(self.iface.mainWindow())
        # self.popupMenu.addAction(self.action1)
        # self.popupMenu.addAction(self.action2)

        # self.toolButton = QToolButton()
        # self.toolButton.setMenu(self.popupMenu)
        # self.toolButton.setDefaultAction(self.action1)
        # self.toolButton.setPopupMode(QToolButton.MenuButtonPopup)

        # self.iface.addToolBarWidget(self.toolButton) 

        # will be set False in run()
        # self.first_start = True
        self.errorMessage = None

        self.apiKey = get_setting("api_key")

        self.configureDialog = ConfigureDialog()
        self.isochronesDialog = IsochronesDialog()
        self.bodDialog = BodDialog()


    def unload(self):
        """Removes the plugin menu item and icon from QGIS GUI."""
        for action in self.actions:
            self.iface.removePluginMenu(plugin_name(), action)
            self.iface.removeToolBarIcon(action)

    def showConfigurationWindow(self):

        if not self.configureDialog.isVisible():
    
            # if self.first_start == True:
            #     self.first_start = False
            #     self.configureDialog = StraTopoConfigureDialog()

            if self.errorMessage is None:
                self.configureDialog.errorLabel.setVisible(False)
            else:
                self.configureDialog.errorLabel.setText(self.errorMessage)
                self.configureDialog.errorLabel.setVisible(True)

            self.configureDialog.apiKeyLineEdit.setText(self.apiKey)
            self.configureDialog.show()
            result = self.configureDialog.exec_()

            # See if OK was pressed
            if result:
                self.apiKey = self.configureDialog.apiKeyLineEdit.text()

                set_setting("api_key", self.apiKey)
        

    
    def showIsochonesWindow(self):
        if not self.isochronesDialog.isVisible():
     
            if self.errorMessage is None:
                self.isochronesDialog.errorLabel.setVisible(False)
            else:
                self.isochronesDialog.errorLabel.setText(self.errorMessage)
                self.isochronesDialog.errorLabel.setVisible(True)


            # TODO: Get list with point layers
            # point_layer_list = []
            # for layer in QgsProject.instance().mapLayers().values():
            #     if type(layer) is QgsVectorLayer:
            #         if layer.geometryType() == QgsWkbTypes.PointGeometry:
            #             point_layer_list.append(layer.name())
            # print(point_layer_list)


            self.isochronesDialog.show()
            result = self.isochronesDialog.exec_()

            # See if OK was pressed
            if result:

                latLineEditText = self.isochronesDialog.latLineEdit.text()
                lonLineEditText = self.isochronesDialog.lonLineEdit.text()
                
                travelTimesCheckBoxFiveChecked = self.isochronesDialog.travelTimesCheckBoxFive.isChecked()
                travelTimesCheckBoxTenChecked = self.isochronesDialog.travelTimesCheckBoxTen.isChecked()
                travelTimesCheckBoxFifteenChecked = self.isochronesDialog.travelTimesCheckBoxFifteen.isChecked()
                travelTimesCheckBoxTwentyChecked = self.isochronesDialog.travelTimesCheckBoxTwenty.isChecked()
                travelTimesCheckBoxTwentyfiveChecked = self.isochronesDialog.travelTimesCheckBoxTwentyfive.isChecked()
                travelTimesCheckBoxThirtyChecked = self.isochronesDialog.travelTimesCheckBoxThirty.isChecked()

                modalityRadioButtonWalkingChecked = self.isochronesDialog.modalityRadioButtonWalking.isChecked()
                modalityRadioButtonBikingChecked = self.isochronesDialog.modalityRadioButtonBiking.isChecked()
                modalityRadioButtonDrivingChecked = self.isochronesDialog.modalityRadioButtonDriving.isChecked()

                # TODO: isochroon voor ieder punt
                # lat = None
                # lon = None
                # selectedLayer = QgsProject.instance().mapLayersByName('test')
                # if not len(selectedLayer) == 1:
                #     print("Wrong lat input")
                #     self.errorMessage = self.tr(u'Error in latitude')
                #     self.showIsochonesWindow()
                #     return

                # if type(selectedLayer[0]) is QgsVectorLayer:
                #     if selectedLayer[0].geometryType() == QgsWkbTypes.PointGeometry:
                #         for feature in selectedLayer[0].getFeatures():
                #             # TODO: check if feature is multipoint
                #             geometry = QgsGeometry.asPoint(feature.geometry())#get the geometry of the feature
                #             point = QgsPointXY(geometry)
                #             lon = point.x() # lan
                #             lat = point.y() # lat


                # Get Lat Long
                if not bool(re.match("^[0-9]{1,3}(?:[\.,][0-9]+)?$", latLineEditText)):
                    print("Wrong lat input")
                    self.errorMessage = self.tr(u'Error in latitude')
                    self.showIsochonesWindow()
                    return
                if not bool(re.match("^[0-9]{1,3}(?:[\.,][0-9]+)?$", lonLineEditText)):
                    print("Wrong lon input")
                    self.errorMessage = self.tr(u'Error in longitude')
                    self.showIsochonesWindow()
                    return

                lat = float(latLineEditText.replace(',', '.')) 
                lon = float(lonLineEditText.replace(',', '.'))

                # Travel Times
                travelTimesList = []

                if travelTimesCheckBoxFiveChecked:
                    travelTimesList.append(5)
                if travelTimesCheckBoxTenChecked:
                    travelTimesList.append(10)
                if travelTimesCheckBoxFifteenChecked:
                    travelTimesList.append(15)
                if travelTimesCheckBoxTwentyChecked:
                    travelTimesList.append(20)
                if travelTimesCheckBoxTwentyfiveChecked:
                    travelTimesList.append(25)
                if travelTimesCheckBoxThirtyChecked:
                    travelTimesList.append(30)

                if len(travelTimesList) == 0:
                    print("At least one travel time should be checked")
                    self.errorMessage = self.tr(u'Error: travel time not selected')
                    self.showIsochonesWindow()
                    return

                # Modality
                modality = ""
                if modalityRadioButtonWalkingChecked:
                    modality = "walk"
                if modalityRadioButtonBikingChecked:
                    modality = "bike"
                if modalityRadioButtonDrivingChecked:
                    modality = "car"

                if modality == "":
                    print("One form of modality should be checked")
                    self.errorMessage = self.tr(u'Error: modality not selected')
                    self.showIsochonesWindow()
                    return

                # Make request
                url = self.apiUrl+"/routing/v1/isochrones"
                requestBody = {
                    "latitude": lat,
                    "longitude": lon,
                    "minutes": travelTimesList,
                    "modality": modality
                }
                headers = {'Content-type': 'application/json', 'Authorization': self.apiKey}
                r = requests.post(url, json=requestBody, headers=headers)

                if r.status_code == 200:
                    self.drawIsochrone(r.json())
                    self.errorMessage = None
                elif r.status_code == 403 or r.status_code == 401:
                    print("Error form server: ", r.status_code)
                    self.errorMessage = self.tr(u'Provide a valid API key and try again')
                    self.showConfigurationWindow()
                else:
                    print("Error form server: ", r.status_code)
                    self.errorMessage = self.tr(u'Server error')
                    self.showIsochonesWindow()
   

    def drawIsochrone(self, geoJson):
        """Draw GeoJson on new layer"""

        if "features" in geoJson and "properties" in geoJson:
            
            newFeatures = []
            for jsonFeature in geoJson["features"]:
                if "geometry" in jsonFeature and "properties" in jsonFeature:
                    if "coordinates" in jsonFeature["geometry"]:

                        newCoordinates = []
                        for coordinates in jsonFeature["geometry"]["coordinates"]:
                            newCoordinatesInnerList = []
                            for coordinate in coordinates:
                                if len(coordinate) == 2:
                                    newCoordinatesInnerList.append(QgsPointXY(coordinate[0], coordinate[1]))
                            newCoordinates.append(newCoordinatesInnerList)

                        newAttributes = ["", "", 0]
                        if "modality" in jsonFeature["properties"]:
                            newAttributes[0] = jsonFeature["properties"]["modality"]
                        if "speed_type" in jsonFeature["properties"]:
                            newAttributes[1] = jsonFeature["properties"]["speed_type"]
                        if "travel_time_minute" in jsonFeature["properties"]:
                            newAttributes[2] = jsonFeature["properties"]["travel_time_minute"]

                        feature = QgsFeature()
                        feature.setGeometry(QgsGeometry.fromPolygonXY(newCoordinates))
                        feature.setAttributes(newAttributes)
                        newFeatures.append(feature)

            if len(newFeatures) > 0 and "latitude" in geoJson["properties"] and "longitude" in geoJson["properties"]:

                # point
                pointLayer =  QgsVectorLayer('Point', 'StartingPoint' , "memory")
                pointDataProvider = pointLayer.dataProvider() 

                pt = QgsFeature()
                point1 = QgsPointXY(geoJson["properties"]["longitude"], geoJson["properties"]["latitude"])
                pt.setGeometry(QgsGeometry.fromPointXY(point1))
                pointDataProvider.addFeatures([pt])

                pointLayer.updateExtents()

                # polygon
                polygonLayer = QgsVectorLayer('Polygon?crs=epsg:4326', 'Isochrones' , 'memory')
                polygonDataProvider = polygonLayer.dataProvider()

                polygonDataProvider.addAttributes([QgsField("modality", QVariant.String), QgsField("speed_type", QVariant.String), QgsField("travel_time_minute", QVariant.Int)])
                polygonLayer.updateFields()

                polygonDataProvider.addFeatures(newFeatures)
                polygonLayer.updateExtents()

                # add layers
                QgsProject.instance().addMapLayers([pointLayer, polygonLayer])


    
    def showBodWindow(self):

        if not self.bodDialog.isVisible():
      
            if self.errorMessage is None:
                self.bodDialog.errorLabel.setVisible(False)
            else:
                self.bodDialog.errorLabel.setText(self.errorMessage)
                self.bodDialog.errorLabel.setVisible(True)

            self.bodDialog.show()
            result = self.bodDialog.exec_()

            # See if OK was pressed
            if result:

                latLineEditText = self.bodDialog.latLineEdit.text()
                lonLineEditText = self.bodDialog.lonLineEdit.text()
                
                bodTypeRadioButtonIntercitytrainChecked = self.bodDialog.bodTypeRadioButtonIntercitytrain.isChecked()
                bodTypeRadioButtonRegionaltrainChecked = self.bodDialog.bodTypeRadioButtonRegionaltrain.isChecked()
                bodTypeRadioButtonBusChecked = self.bodDialog.bodTypeRadioButtonBus.isChecked()
                bodTypeRadioButtonTramChecked = self.bodDialog.bodTypeRadioButtonTram.isChecked()
                bodTypeRadioButtonMetroChecked = self.bodDialog.bodTypeRadioButtonMetro.isChecked()
                bodTypeRadioButtonCitycenterChecked = self.bodDialog.bodTypeRadioButtonCitycenter.isChecked()
                bodTypeRadioButtonShoppingcenterChecked = self.bodDialog.bodTypeRadioButtonShoppingcenter.isChecked()
                bodTypeRadioButtonSmallshopChecked = self.bodDialog.bodTypeRadioButtonSmallshop.isChecked()
                bodTypeRadioButtonElementaryschoolChecked = self.bodDialog.bodTypeRadioButtonElementaryschool.isChecked()
                bodTypeRadioButtonHighschoolChecked = self.bodDialog.bodTypeRadioButtonHighschool.isChecked()
                bodTypeRadioButtonUniversityChecked = self.bodDialog.bodTypeRadioButtonUniversity.isChecked()
                bodTypeRadioButtonCommutingChecked = self.bodDialog.bodTypeRadioButtonCommuting.isChecked()

                travelTimeLineEditText = self.bodDialog.travelTimeLineEdit.text()

                # Get Lat Long
                if not bool(re.match("^[0-9]{1,3}(?:[\.,][0-9]+)?$", latLineEditText)):
                    print("Wrong lat input")
                    self.errorMessage = self.tr(u'Error in latitude')
                    self.showBodWindow()
                    return
                if not bool(re.match("^[0-9]{1,3}(?:[\.,][0-9]+)?$", lonLineEditText)):
                    print("Wrong lon input")
                    self.errorMessage = self.tr(u'Error in longitude')
                    self.showBodWindow()
                    return

                lat = float(latLineEditText.replace(',', '.')) 
                lon = float(lonLineEditText.replace(',', '.'))

                # bod_type
                bod_type = ""
                if bodTypeRadioButtonIntercitytrainChecked:
                    bod_type = "intercitytrain"
                if bodTypeRadioButtonRegionaltrainChecked:
                    bod_type = "regionaltrain"
                if bodTypeRadioButtonBusChecked:
                    bod_type = "bus"
                if bodTypeRadioButtonTramChecked:
                    bod_type = "tram"
                if bodTypeRadioButtonMetroChecked:
                    bod_type = "metro"
                if bodTypeRadioButtonCitycenterChecked:
                    bod_type = "citycenter"
                if bodTypeRadioButtonShoppingcenterChecked:
                    bod_type = "shoppingcenter"
                if bodTypeRadioButtonSmallshopChecked:
                    bod_type = "smallshop"
                if bodTypeRadioButtonElementaryschoolChecked:
                    bod_type = "elementaryschool"
                if bodTypeRadioButtonHighschoolChecked:
                    bod_type = "hogeschool"
                if bodTypeRadioButtonUniversityChecked:
                    bod_type = "university"
                if bodTypeRadioButtonCommutingChecked:
                    bod_type = "commuting"
                
                if bod_type == "":
                    print("One form of modality should be checked")
                    self.errorMessage = self.tr(u'Error: modality not selected')
                    self.showBodWindow()
                    return

                # travelTime 
                if not bool(re.match("^[0-9]+$", travelTimeLineEditText)):
                    print("Wrong travel time input")
                    self.errorMessage = self.tr(u'Travel time should be a number')
                    self.showBodWindow()
                    return

                minutes = int(travelTimeLineEditText)
                if minutes < 1 or minutes > 60:
                    print("Travel time should be between 1 and 60")
                    self.errorMessage = self.tr(u'Travel time should be between 1 and 60')
                    self.showBodWindow()
                    return
                
                # Make request
                url = self.apiUrl+"/routing/v1/bod"
                requestBody = {
                    "latitude": lat,
                    "longitude": lon,
                    "bod_type": bod_type,
                    "minutes": minutes
                }
                headers = {'Content-type': 'application/json', 'Authorization': self.apiKey}
                r = requests.post(url, json=requestBody, headers=headers)

                if r.status_code == 200:
                    self.drawBod(r.json())
                    self.errorMessage = None
                elif r.status_code == 403 or r.status_code == 401:
                    print("Error form server: ", r.status_code)
                    self.errorMessage = self.tr(u'Provide a valid API key and try again')
                    self.showConfigurationWindow()
                else:
                    print("Error form server: ", r.status_code)
                    self.errorMessage = self.tr(u'Server error')
                    self.showBodWindow()


    def drawBod(self, geoJson):

        if "aantal_bag_objecten" in geoJson and \
            "bod_score" in geoJson and \
            "bod_type" in geoJson and \
            "latitude" in geoJson and \
            "longitude" in geoJson and \
            "minutes" in geoJson and \
            "nod_score" in geoJson and \
            "sum_gem_hh_gr" in geoJson:

            # point
            pointLayer =  QgsVectorLayer('Point', 'BOD_Point' , "memory")
            pointDataProvider = pointLayer.dataProvider() 
            pointDataProvider.addAttributes([QgsField("aantal_bag_objecten", QVariant.Int), QgsField("bod_score", QVariant.Double), QgsField("bod_type", QVariant.String), QgsField("minutes", QVariant.Int), QgsField("nod_score", QVariant.Double), QgsField("sum_gem_hh_gr", QVariant.Double)])
            pointLayer.updateFields()

            feature = QgsFeature()
            point1 = QgsPointXY(geoJson["longitude"], geoJson["latitude"])
            feature.setGeometry(QgsGeometry.fromPointXY(point1))

            attributes = [0, None, "", 0, None, None]
            attributes[0] = geoJson["aantal_bag_objecten"]
            attributes[1] = geoJson["bod_score"]
            attributes[2] = geoJson["bod_type"]
            attributes[3] = geoJson["minutes"]
            attributes[4] = geoJson["nod_score"]
            attributes[5] = geoJson["sum_gem_hh_gr"]
             
            feature.setAttributes(attributes)
            pointDataProvider.addFeatures([feature])
            pointLayer.updateExtents()


            # add layers
            QgsProject.instance().addMapLayers([pointLayer])